#version 330

#define M_PI 3.14159265359

#define M_GAMMA 2.2

// Position of the fragment in camera space:
in vec3 position;

// UV coordinate:
in vec2 uv;

// Normal in camera space:
in vec3 normal;

// Tangent in camera space:
in vec3 tangent;

// Tangent-to-Camera Space transformation matrix
in mat3 TBN;

// Define ambient light intensity:
const vec3 ambientLight = vec3(0.05,0.05,0.05);

// Defines whether pbr lighting should be used (or phong lighting):
uniform bool usePBRLighting;

// Base Color Texture:
uniform bool hasBaseColorTexture;
uniform sampler2D baseColorTexture;

// Metallic Texture:
uniform bool hasMetallicTexture;
uniform sampler2D metallicTexture;

// Roughness Texture:
uniform bool hasRoughnessTexture;
uniform sampler2D roughnessTexture;

// Normal Texture:
uniform bool hasNormalTexture;
uniform sampler2D normalTexture;

// Ambient Occlusion Texture:
uniform bool hasAmbientOcclusionTexture;
uniform sampler2D ambientOcclusionTexture;

// Default shininess value (when no metallicTexture is
// available, use this!):
uniform float phongDefaultShininess;

// Default metallic value (when no metallicTexture is
// available, use this!):
uniform float pbrDefaultMetallic;

// Default roughtness value (when no roughnessTexture
// is available, use this!):
uniform float pbrDefaultRoughness;

// Defines all the data a light source needs:
struct Light {
    vec3 position;
    vec3 intensity;         // <- for pbr lighting
    vec3 phong_diffuse;     // <- for phong lighting
    vec3 phong_specular;    // <- for phong lighting
};

// Defines 100 lights:
uniform Light lights[100];

// Only use as many lights as wanted:
uniform int lightNum;

// Should the PBR path be used?
uniform bool usePBR;

// The output of this fragment shader:
out vec4 fragment_color;

/**
 * Returns the fragment color using the phong lighting.
 */
vec3 getPhongLighting(){
    vec3 kD = vec3(1.0, 1.0, 1.0);

    if(hasBaseColorTexture){
        kD *= texture(baseColorTexture, uv).rgb;
    }

    vec3 iOut = kD * ambientLight;

    for(int i = 0; i < lightNum; ++i){
        Light light = lights[i];
        vec3 l = light.position - position;
        float lightDist = length(l);
        l /= lightDist;

        vec3 n = normalize(normal);

        if(hasNormalTexture){
            n = texture(normalTexture, uv).xyz * 2.0 - 1.0;
            n = normalize(TBN * n);
        }

        float nDotL = dot(n, l);

        vec3 e = normalize(-position);
        vec3 h = (l + e) / length(l + e);

        if(nDotL < 0.0)
           continue;

        float intensity = 1.0 / (lightDist * lightDist);
        iOut += kD * light.phong_diffuse * nDotL * intensity;
        iOut += kD * light.phong_specular * max(0, pow(dot(h, n), phongDefaultShininess)) * intensity;
    }

    iOut = pow(iOut, vec3(1.0/M_GAMMA));
    return iOut;
}




/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 *
 * Schlicks approximation for fresnel factor:
 */

vec3 fresnel(float cosTheta, vec3 r0)
{
    return r0 + (1.0 - r0) * pow(1.0 - cosTheta, 5.0);
}

float fresnel(float cosTheta, float r0)
{
    return r0 + (1.0 - r0) * pow(1.0 - cosTheta, 5.0);
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 *
 * Trowbridge-Reitz GGX (Normal Distribution Function):
 */

float D_GGX(vec3 normal, vec3 halfVec, float pRoughness)
{
    float a = pRoughness * pRoughness;
    float a2 = a * a;
    float NdH = max(dot(normal, halfVec), 0.0);
    float NdH2 = NdH*NdH;
    float p = (NdH2 * (a2 - 1.0) + 1.0);

    return a2 / (M_PI * p * p);
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 *
 * Schlicks approximation of the smith equation:
 */

float G_Schlick(float NdV, float pRoughness)
{
    float r = (pRoughness + 1.0) / 2;
    float b = (r*r);

    return NdV / (NdV * (1.0 - b) + b);
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 *
 * Smith equation (G is divided into two components: light & view):
 */

float G_Smith(vec3 N, vec3 V, vec3 L, float pRoughness)
{
    float g1  = G_Schlick(max(dot(N, V), 0.0), pRoughness);
    float g2  = G_Schlick(max(dot(N, L), 0.0), pRoughness);

    return g1 * g2;
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 */

vec3 MicroFacetBRDF(vec3 normal, vec3 halfVec, vec3 toCamera, vec3 toLight, vec3 F, float pRoughness){
    // Berechne NDF und G Term:
    float NDF = D_GGX(normal, halfVec, pRoughness);
    float G = G_Smith(normal, toCamera, toLight, pRoughness);

    // Berechne die Cook-Torrance BRDF:
    vec3 nominator = NDF * G * F;
    float denominator = 4 * max(dot(normal, toCamera), 0.0)
            * max(dot(normal, toLight), 0.0) + 0.0001;

    return nominator / denominator;
}

/**
 * Returns the fragment color using the pbr lighting.
 *
 * TODO: Implement this function!
 */


vec3 getPBRLighting(){
    // Vectors to fragment and to camera in view space:
    vec3 toFragment = normalize(position);
    vec3 toCamera = -toFragment;

    vec3 rBaseColor = vec3(1.0, 1.0, 1.0);
    float rMetallic = pbrDefaultMetallic;
    float rRoughness = pbrDefaultRoughness;
    vec3 rNormal = normal;

    if(hasBaseColorTexture)
        rBaseColor *= texture(baseColorTexture, uv).rgb;

    if(hasMetallicTexture)
        rMetallic = texture(metallicTexture, uv).r;

    if(hasRoughnessTexture)
        rRoughness = texture(roughnessTexture, uv).r;

    if(hasNormalTexture){
        rNormal = texture(normalTexture, uv).xyz * 2.0 - 1.0;
        rNormal = normalize(TBN * rNormal);
    }

    // Initialize colors:
    vec3 totalDiffuse = ambientLight * rBaseColor;
    vec3 totalSpecular = vec3(0.0,0.0,0.0);

    // Reflection coefficient if view dir is perpendicular to face:
    vec3 r0 = mix(vec3(0.04), rBaseColor, rMetallic).rgb;

    // Add all light sources:
    for(int i=0; i < lightNum ; ++i){
        vec3 lightPosition = lights[i].position;

        // Vector from fragment to light:
        vec3 toLight = normalize(lightPosition - position);

        // Attenuation:
        float dist = distance(lightPosition, position);

        // Calculate light intensity:
        vec3 radiance = lights[i].intensity / (dist * dist);

        float NdotL = max(dot(rNormal,toLight),0.0);

        // Half Vector (fragment->light + fragment->camera):
        vec3 halfVec = normalize(toLight + toCamera);

        vec3 F = fresnel(max(dot(halfVec, toCamera), 0.0), r0);

        vec3 specular = MicroFacetBRDF(rNormal, halfVec, toCamera, toLight, F, rRoughness);

        vec3 kD = (vec3(1.0) - F) * (1.0 - rMetallic);

        totalDiffuse += max((kD * rBaseColor / M_PI) * radiance * NdotL, vec3(0.0,0.0,0.0));
        totalSpecular += specular * radiance * NdotL;
    }

    totalDiffuse = pow(totalDiffuse, vec3(1.0/M_GAMMA));
    totalSpecular = pow(totalSpecular, vec3(1.0/M_GAMMA));

    return totalDiffuse + totalSpecular;
}

/**
 * The main function of the fragment shader, which is executed
 * for every fragment.
 */
void main()
{
    if(usePBR)
        fragment_color = vec4(getPBRLighting(), 1.0);
    else
        fragment_color = vec4(getPhongLighting(), 1.0);
}
